"""Ensure that experiment pre-execution maps experiments as expected."""

# tests spec below
"""Ensure that an aggregate stage map correctly has the record references that went into it."""

"""Ensure that cache detection works for regular stage maps."""
