"""This file is simply a manual way into the experiment script
from curifactory - this allows running it from inside an IDE
such as pycharm, in order to get debugging."""

from curifactory import experiment

experiment.main()
